// genemap.H
// Bokhari Gene Mapping Object header file
// super of the gene object
// does the problem specific handling

#include "gene.H"

#define MINIMAL_MUTATION 10000
#define MINIMAL_CROSS      100

extern int CROSS_TYPE;
extern int FITNESS_TYPE;
extern int MIX_FACTOR;
extern float MUT_FACTOR;
extern float CROSS_RATE;

class Genemap : public Gene
{
public:
  Meshmap *mesh;          // pointer to problem defintion

  // Construct and Destruct
  Genemap(void);
  ~Genemap(void);

  // Initialize new Genemap from source
  void operator=(const Genemap&);
  int Randomize(Meshmap *);
  int Cross(Genemap&, Genemap&, Meshmap *);
  int CrossPMX(Genemap&, Genemap&);
  int CrossOX(Genemap&, Genemap&);

  // Internal information and modification
  double Fitness();
  int Mutate(void);

  // Information
  //int Get_Map(int) const;        // get a particular mapping
  //inline double Get_Fit(void) const
  //{ return fitness; }     // return the fitness
  //inline int Get_Size(void) const
  //{ return size; }        // get the size of the mesh (nodes)
};

// Constructs a Gene
Genemap::Genemap(void)
{
  size = 0;
  edges = 0;
  map = (int *) NULL;
  fitness = 0.0;
}

// Deconstructs a Gene
Genemap::~Genemap(void)
{
  delete map;
}

// Copy operator
void Genemap::operator=(const Genemap& from)
{
  int count;

  // copy over basic values
  size = from.size;
  fitness = from.fitness;
  mesh = from.mesh;

  map = new int[size];        // find space for the map

  // copy over the mapping
  for(count=0;(count<size);count++)
    map[count] = from.map[count];
}

// Initializes and randomizes a Gene for a given problem size
int Genemap::Randomize(Meshmap *targ_mesh)
{
  mesh = targ_mesh;

  // Construct the internal map array
  size = mesh->Get_Size();
  map = new int[size];

  // temp holder for mach_size... (this really needs to be redesigned
  int mach_size;
  mach_size = mesh->Get_Mach_Size();

  // Create a random mapping of the problem elements to the PEs
  int count;
  for(count=0;(count < size);count++)
    map[count] = ((int) random()) % mesh->Get_Mach_Size();

  return(1);
}

// Fitness function
// determines the fitness of a particular mapping
// this uses both a loading and a communication system given equal weight
double Genemap::Fitness()
{
  int cur_node,cur_edge;
  int node,target;
  double maxfit;
  double comm_fit = 0.0;
  double load_fit = 0.0;

  // commonly used but from another object
  int mach_size, mach_side;
  mach_size = mesh->Get_Mach_Size();
  mach_side = mesh->Get_Mach_Side();

  // Communication Fitness Section
  int x1,x2,y1,y2,dx,dy,temp;
  for(cur_node=0;(cur_node < size);cur_node++)
    for(cur_edge=0,node=Get_Map(cur_node);
	(cur_edge < mesh->Get_Connect());cur_edge++)
      {
	if (mesh->Get_Edge(cur_node,cur_edge) != -1)
	  {
	    target = Get_Map(mesh->Get_Edge(cur_node,cur_edge));
	    
	    // cout << "node: "<<node<< " target: "<<target<< " ";

	    // Modified Manhattan Distance
	    x1 = node%mach_side;
	    x2 = target%mach_side;
	    dx = abs(x1 - x2);
	    temp = ((x1>x2) ? x2+abs(mach_side-x1) :
		    x1+abs(mach_side-x2));
	    dx = ((dx>temp) ? temp : dx);
	    
	    y1 = node/(mach_side);
	    y2 = target/(mach_side);
	    dy = abs(y1 - y2);
	    temp = ((y1>y2) ? y2 + abs((mach_side)-y1) :
		    y1 + abs((mach_side)-y2));
	    dy = ((dy>temp) ? temp : dy);

	    // cout <<"x1 "<<x1<<" x2 "<<x2<<" y1 "<<y1<<" y2 "<<y2;
	    // cout <<"dx "<<dx<<" dy "<<dy<<endl;
	    comm_fit += (double) ((dx>dy) ? dy : dx) + abs(dx-dy);
	  }
      }
  comm_fit = comm_fit / 2.0;
  maxfit = (double) (mesh->Get_Max_Edges() * (mach_side/2));
  comm_fit = 1.0 - (comm_fit/maxfit);

  // cout << "comm: " << comm_fit << " max " << maxfit << endl;

  // Loading Fitness Section
  double avg_load, run_total;
  double tmp_load;
  int *map_total;

  // Compile the load on each processor
  map_total = new int[mach_size];
  // Intialize temporary totalling array
  for(target=0;(target < mach_size);target++)
    map_total[target] = 0;
  // increment the appropos map_total for each mapping of a node
  for(node =0;(node < size);node++)
    map_total[Get_Map(node)]++;

  // figure the average load on a given PE
  avg_load = (double) size / (double) mach_size;

  // figure the variance of each PE's load, keep a running total
  for(target=0, run_total=0.0;(target < mach_size);target++)
    {
      tmp_load = (double) map_total[target];
      run_total += (tmp_load*tmp_load) - (avg_load*avg_load);
    }
  load_fit = run_total / (double) mach_size;
  // Figure worst case
  tmp_load = (double) (size*size) - ((double) mach_size - 1.0) * avg_load;
  load_fit = (tmp_load - load_fit)/tmp_load;
  delete map_total;   // reclaim memory of map loading array

  // fitness is equal weighting of load fitness and communication fitness
  fitness = (comm_fit + load_fit) / 2.0;

  // cout <<fitness <<"= comm:" <<comm_fit <<" & load:" <<load_fit <<endl;
  // Info();

  return(fitness);
}

int Genemap::Cross(Genemap& mom, Genemap& dad, Meshmap *targ_mesh)
{
  // Construct the internal map array
  mesh = targ_mesh;
  size = mesh->Get_Size();
  map = new int[size];

  int val=0;
  switch(CROSS_TYPE)
    {
    case PMX:
      val = CrossPMX(mom, dad);
      break;
    case OX:
      val = CrossOX(mom, dad);
      break;
    default:
      val = CrossPMX(mom, dad);
      break;
    }

  // will we cross-over... yes we\'ve already done the effort, but this
  // keeps my twin markers going

  int chance,count;
  chance = ((int) random()) % MINIMAL_CROSS;
  if (chance > (MINIMAL_CROSS * CROSS_RATE))
    {
      // if no cross over, then clone mom into the next generation
      fitness = mom.Get_Fit();
      for(count=0;(count<size);count++)
	map[count] = mom.map[count];
      val = 1;
    }

  return(val);
}

// Cross Breeding function by order crossover operator (Goldberg[1989])
// returns 1 if the gene is cross bred
int Genemap::CrossOX(Genemap& mom, Genemap& dad)
{
  // The twins and the twin born flag
  int john[size], jane[size];

  int start,stop,dist,count;
  int work_john[size],work_jane[size];

  for(count=0;(count<size);count++)
    {
      work_john[count] = mom.map[count];
      work_jane[count] = dad.map[count];
    }

  // choose two random crossover points
  start = ((int) random()) % size;
  stop = ((int) random()) % size;
  if (start>stop)
    {
      dist = stop;
      stop = start;
      start = dist;
    }
  dist = (stop - start)+1;

  // cut out the splice of john and jane
  int splice;
  int cut_john[dist], cut_jane[dist];
  for(splice=0,count=start;(count<=stop);count++,splice++)
    {
      cut_john[splice] = work_john[count];
      cut_jane[splice] = work_jane[count];
    }
  for(count=0;(count<size);count++)
    for(splice=0;(splice<dist);splice++)
      {
	if (work_john[count] == cut_jane[splice])
	  work_john[count] = -1;
	if (work_jane[count] == cut_john[splice])
	  work_jane[count] = -1;
      }


  // slide the holes together at the start point
  for(count=start;(count<=stop);count++)
    john[count] = jane[count] = -1;
  for(splice=0,count=0;(count<size);count++)
    {
      if (work_john[count] != -1)
	{
	  john[splice] = work_john[count];
	  splice++;
	  splice=((splice==start) ? (stop+1) : splice);
	}
    }
  for(splice=0,count=0;(count<size);count++)
    {
      if (work_jane[count] != -1)
	{
	  jane[splice] = work_jane[count];
	  splice++;
	  splice=((splice==start) ? (stop+1) : splice);
	}
    }

  // patch the splice holes
  for(count=start;(count<=stop);count++)
    {
      john[count] = cut_jane[(count-start)];
      jane[count] = cut_john[(count-start)];
    }

  // set the current child to the john twin
  for(count=0;(count<size);count++)
      map[count] = john[count];

  return(1);
}

// Cross Breeding function by partial matching
// returns 1 if the gene is cross bred
int Genemap::CrossPMX(Genemap& mom, Genemap& dad)
{
  int start, stop, count, mach_size;
  // Initialize the child gene to a null map
  for(count=0;(count<size);count++)
    map[count]=-1;
  
  // make a local version of the mesh->mach_size
  mach_size = mesh->Get_Mach_Size();

  // Choose a start and stop point for the twist
  start = ((int) random()) % size;
  stop = ((int) random()) % size;
  
  //cout << "\nstart:"<<start<<" stop:"<<stop<<endl;
  
  // replicate mom over to the child
  for(count=start;(count!=((stop+1)%size));count = (count+1)%size)
    map[count] = mom.map[count];
  
  // replicate dad over to the child
  for(count=(stop+1)%size;(count!=start);count = (count+1)%size)
    map[count] = dad.map[count];
  
  // Where there is a duplication, replace with other parent
  int *fill,temp,done,monitor=0;
  fill = new int[mach_size];
  do
    {
      done = 1;
      for(count=0;(count<mach_size);count++)
	fill[count] = -1;
      for(count=0;(count<size);count++)
	{
	  temp = map[count];
	  if (fill[temp] == -1)
	    fill[temp] = count;
	  else
	    {
	      if (monitor>size)
		{
		  // We musta hit a loop, try to break out
		  map[count] = ((int) random())%size;
		  monitor = size/2;
		  //cerr << "Monitor Break!\n";
		}
	      else
		{
		  // gotta swap one and recheck
		  if (map[count] == mom.map[count])
		    map[count] = dad.map[count];
		  else
		    map[count] = mom.map[count];
		  done = 0;
		  monitor++;
		}
	    }
	}
    } while (!done);
  delete fill;
  return(1);
}

// Mutation function, returns 1 if mutated
int Genemap::Mutate()
{
  int chance, a, b, temp;
  chance = ((int) random()) % MINIMAL_MUTATION;
  if (chance < (MINIMAL_MUTATION * MUT_FACTOR))
    {
      a = ((int) random()) % size;
      b = ((int) random()) % size;
      temp = map[a];
      map[a] = map[b];
      map[b] = temp;
      return(1);
    }
  else
    return(0);
}
